import csv
import os
from pathlib import Path

import pytest

files = sorted(Path(os.environ["CATALOG_DIR"]).glob("*.csv"))


def catalog_name(catalog) -> str:
    return catalog.name


class TestAllCatalogs:
    @pytest.fixture(params=files, ids=catalog_name)
    def catalog(self, request):
        yield request.param

    def test_non_zero_cost(self, catalog):
        reader = csv.DictReader(catalog.open())
        for row in reader:
            assert float(row["price"]) != pytest.approx(0), str(row)



import copy
import logging
from collections import defaultdict
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import requests

from gpuhunt._internal.constraints import KNOWN_GPUS
from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
bundles_url = "https://console.vast.ai/api/v0/bundles/"
kilo = 1000
Operators = Literal["lt", "lte", "eq", "gte", "gt"]
FilterValue = Union[int, float, str, bool]


class VastAIProvider(AbstractProvider):
    NAME = "vastai"

    def __init__(self, extra_filters: Optional[Dict[str, Dict[Operators, FilterValue]]] = None):
        self.extra_filters = extra_filters

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> List[RawCatalogItem]:
        filters: Dict[str, Any] = self.make_filters(query_filter or QueryFilter())
        if self.extra_filters:
            for key, constraints in self.extra_filters.items():
                for op, value in constraints.items():
                    filters[key][op] = value
        filters["rentable"]["eq"] = True
        filters["rented"]["eq"] = False
        filters["order"] = [["score", "desc"]]
        resp = requests.post(bundles_url, json=filters)
        resp.raise_for_status()
        data = resp.json()

        instance_offers = []
        for offer in data["offers"]:
            if not self.satisfies_filters(offer, filters):
                logger.warning("Offer %s does not satisfy filters", offer["id"])
                continue
            gpu_name = get_gpu_name(offer["gpu_name"])
            gpu_memory = normalize_gpu_memory(gpu_name, offer["gpu_ram"])
            ondemand_offer = RawCatalogItem(
                instance_name=str(offer["id"]),
                location=get_location(offer["geolocation"]),
                # storage_cost is $/gb/month
                price=round(
                    offer["dph_base"]
                    + (query_filter.min_disk_size or 0) * offer["storage_cost"] / 30 / 24,
                    5,
                ),
                cpu=int(offer["cpu_cores_effective"]),
                memory=float(
                    int(
                        offer["cpu_ram"] * offer["cpu_cores_effective"] / offer["cpu_cores"] / kilo
                    )
                ),
                gpu_count=offer["num_gpus"],
                gpu_name=gpu_name,
                gpu_memory=float(gpu_memory),
                spot=False,
                disk_size=offer["disk_space"],
            )
            instance_offers.append(ondemand_offer)

            spot_offer = copy.deepcopy(ondemand_offer)
            spot_offer.price = round(offer["min_bid"], 5)
            spot_offer.spot = True
            instance_offers.append(spot_offer)
        return instance_offers

    @staticmethod
    def make_filters(q: QueryFilter) -> Dict[str, Dict[Operators, FilterValue]]:
        filters = defaultdict(dict)
        if q.min_cpu is not None:
            filters["cpu_cores"]["gte"] = q.min_cpu
        if q.max_cpu is not None:
            filters["cpu_cores"]["lte"] = q.max_cpu
        if q.min_memory is not None:
            filters["cpu_ram"]["gte"] = q.min_memory * kilo
        if q.max_memory is not None:
            filters["cpu_ram"]["lte"] = q.max_memory * kilo
        if q.min_gpu_count is not None:
            filters["num_gpus"]["gte"] = q.min_gpu_count
        if q.max_gpu_count is not None:
            filters["num_gpus"]["lte"] = q.max_gpu_count
        # We cannot reliably filter by GPU memory, because it is not the same for a specific GPU model
        if q.min_disk_size is not None:
            filters["disk_space"]["gte"] = q.min_disk_size
        if q.max_disk_size is not None:
            filters["disk_space"]["lte"] = q.max_disk_size
        if q.min_price is not None:
            filters["dph_total"]["gte"] = q.min_price
        if q.max_price is not None:
            filters["dph_total"]["lte"] = q.max_price
        # TODO(egor-s): add compute capability info for all GPUs
        if q.min_compute_capability is not None:
            filters["compute_capability"]["gte"] = compute_cap(q.min_compute_capability)
        if q.max_compute_capability is not None:
            filters["compute_capability"]["lte"] = compute_cap(q.max_compute_capability)
        return filters

    @staticmethod
    def satisfies_filters(offer: dict, filters: Dict[str, Dict[Operators, FilterValue]]) -> bool:
        for key in filters:
            if key not in offer:
                continue
            for op, value in filters[key].items():
                if op == "lt" and offer[key] >= value:
                    return False
                if op == "lte" and offer[key] > value:
                    return False
                if op == "eq" and offer[key] != value:
                    return False
                if op == "gte" and offer[key] < value:
                    return False
                if op == "gt" and offer[key] <= value:
                    return False
        return True


def get_gpu_name(gpu_name: str) -> str:
    gpu_name = gpu_name.replace("RTX A", "A").replace("Tesla ", "").replace("Q ", "")
    if gpu_name.startswith("A100 "):
        return "A100"
    return gpu_name.replace(" ", "")


def normalize_gpu_memory(gpu_name: str, memory_mib: float) -> int:
    known_memory = [gpu.memory for gpu in KNOWN_GPUS if gpu.name == gpu_name]
    if known_memory:
        # return the closest known value
        return min(known_memory, key=lambda x: abs(x - memory_mib / kilo))
    return int(memory_mib / kilo)


def get_location(location: Optional[str]) -> str:
    if location is None:
        return ""
    try:
        city, country = location.replace(", ", ",").split(",")
        location = f"{country}-{city}"
    except ValueError:
        pass
    return location.lower().replace(" ", "")


def compute_cap(cc: Tuple[int, int]) -> str:
    """
    >>> compute_cap((7, 0))
    '700'
    >>> compute_cap((7, 5))
    '750'
    """
    major, minor = cc
    return f"{major}{str(minor).ljust(2, '0')}"



from typing import List

import pytest

from gpuhunt import CatalogItem, QueryFilter
from gpuhunt._internal.constraints import matches


@pytest.fixture
def item() -> CatalogItem:
    return CatalogItem(
        instance_name="large",
        location="us-east-1",
        price=1.2,
        cpu=16,
        memory=64.0,
        gpu_count=1,
        gpu_name="A100",
        gpu_memory=40.0,
        spot=False,
        provider="aws",
        disk_size=None,
    )


@pytest.fixture
def cpu_items() -> List[CatalogItem]:
    datacrunch = CatalogItem(
        instance_name="CPU.120V.480G",
        location="ICE-01",
        price=3.0,
        cpu=120,
        memory=480.0,
        gpu_count=0,
        gpu_name=None,
        gpu_memory=0.0,
        spot=False,
        provider="datacrunch",
        disk_size=None,
    )
    nebius = CatalogItem(
        instance_name="standard-v2",
        location="eu-north1-c",
        price=1.4016,
        cpu=48,
        memory=288.0,
        gpu_count=0,
        gpu_name=None,
        gpu_memory=None,
        spot=False,
        provider="nebius",
        disk_size=None,
    )
    return [datacrunch, nebius]


class TestMatches:
    def test_empty(self, item: CatalogItem):
        assert matches(item, QueryFilter())

    def test_cpu(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_cpu=16))
        assert matches(item, QueryFilter(max_cpu=16))
        assert not matches(item, QueryFilter(min_cpu=32))
        assert not matches(item, QueryFilter(max_cpu=8))

    def test_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_memory=64.0))
        assert matches(item, QueryFilter(max_memory=64.0))
        assert not matches(item, QueryFilter(min_memory=128.0))
        assert not matches(item, QueryFilter(max_memory=32.0))

    def test_gpu_count(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_gpu_count=1))
        assert matches(item, QueryFilter(max_gpu_count=1))
        assert not matches(item, QueryFilter(min_gpu_count=2))
        assert not matches(item, QueryFilter(max_gpu_count=0))

    def test_gpu_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_gpu_memory=40.0))
        assert matches(item, QueryFilter(max_gpu_memory=40.0))
        assert not matches(item, QueryFilter(min_gpu_memory=80.0))
        assert not matches(item, QueryFilter(max_gpu_memory=20.0))

    def test_gpu_name(self, item: CatalogItem):
        assert matches(item, QueryFilter(gpu_name=["A100"]))
        assert not matches(item, QueryFilter(gpu_name=["A10"]))

    def test_total_gpu_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_total_gpu_memory=40.0))
        assert matches(item, QueryFilter(max_total_gpu_memory=40.0))
        assert not matches(item, QueryFilter(min_total_gpu_memory=80.0))
        assert not matches(item, QueryFilter(max_total_gpu_memory=20.0))

    def test_price(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_price=1.2))
        assert matches(item, QueryFilter(max_price=1.2))
        assert not matches(item, QueryFilter(min_price=1.3))
        assert not matches(item, QueryFilter(max_price=1.1))

    def test_spot(self, item: CatalogItem):
        assert matches(item, QueryFilter(spot=False))
        assert not matches(item, QueryFilter(spot=True))

    def test_compute_capability(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_compute_capability=(8, 0)))
        assert matches(item, QueryFilter(max_compute_capability=(8, 0)))
        assert not matches(item, QueryFilter(min_compute_capability=(8, 1)))
        assert not matches(item, QueryFilter(max_compute_capability=(7, 9)))

    def test_ti_gpu(self):
        item = CatalogItem(
            instance_name="large",
            location="us-east-1",
            price=1.2,
            cpu=16,
            memory=64.0,
            gpu_count=1,
            gpu_name="RTX3060Ti",  # case-sensitive
            gpu_memory=8.0,
            spot=False,
            provider="aws",
            disk_size=None,
        )
        assert matches(item, QueryFilter(gpu_name=["RTX3060TI"]))

    def test_provider(self, cpu_items):
        assert matches(cpu_items[0], QueryFilter(provider=["datacrunch"]))
        assert matches(cpu_items[1], QueryFilter(provider=["nebius"]))



from pathlib import Path

import pytest


@pytest.fixture
def data(catalog_dir: Path) -> str:
    return (catalog_dir / "nebius.csv").read_text()


class TestNebiusCatalog:
    def test_zone_presented(self, data: str):
        assert ",eu-north1-c," in data

    def test_gpu_presented(self, data: str):
        gpus = [
            "A100",
            "L4",
            "L40",
            "H100",
        ]
        assert all(f",{i}," in data for i in gpus)

    def test_h100_platforms(self, data: str):
        platforms = [
            "gpu-h100",
            "gpu-h100-b",
            "standard-v3-h100-pcie",
        ]
        assert all(f"\n{i}," in data for i in platforms)

    def test_no_spots(self, data: str):
        assert ",True\n" not in data



from typing import Optional, Tuple, TypeVar, Union

from gpuhunt._internal.models import CatalogItem, GPUInfo, QueryFilter

Comparable = TypeVar("Comparable", bound=Union[int, float, Tuple[int, int]])


def is_between(value: Comparable, left: Optional[Comparable], right: Optional[Comparable]) -> bool:
    if is_below(value, left) or is_above(value, right):
        return False
    return True


def is_below(value: Comparable, limit: Optional[Comparable]) -> bool:
    if limit is not None and value < limit:
        return True
    return False


def is_above(value: Comparable, limit: Optional[Comparable]) -> bool:
    if limit is not None and value > limit:
        return True
    return False


def matches(i: CatalogItem, q: QueryFilter) -> bool:
    """
    Check if the catalog item matches the filters

    Args:
        i: catalog item
        q: filters

    Returns:
        whether the catalog item matches the filters
    """
    if q.provider is not None and i.provider.lower() not in q.provider:
        return False
    if not is_between(i.cpu, q.min_cpu, q.max_cpu):
        return False
    if not is_between(i.memory, q.min_memory, q.max_memory):
        return False
    if not is_between(i.gpu_count, q.min_gpu_count, q.max_gpu_count):
        return False
    if q.gpu_name is not None:
        if i.gpu_name is None:
            return False
        if i.gpu_name.lower() not in q.gpu_name:
            return False
    if q.min_compute_capability is not None or q.max_compute_capability is not None:
        cc = get_compute_capability(i.gpu_name)
        if not cc or not is_between(cc, q.min_compute_capability, q.max_compute_capability):
            return False
    if not is_between(i.gpu_memory if i.gpu_count > 0 else 0, q.min_gpu_memory, q.max_gpu_memory):
        return False
    if not is_between(
        (i.gpu_count * i.gpu_memory) if i.gpu_count > 0 else 0,
        q.min_total_gpu_memory,
        q.max_total_gpu_memory,
    ):
        return False
    if i.disk_size is not None:
        if not is_between(i.disk_size, q.min_disk_size, q.max_disk_size):
            return False
    if not is_between(i.price, q.min_price, q.max_price):
        return False
    if q.spot is not None and i.spot != q.spot:
        return False
    return True


def get_compute_capability(gpu_name: str) -> Optional[Tuple[int, int]]:
    for gpu in KNOWN_GPUS:
        if gpu.name.lower() == gpu_name.lower():
            return gpu.compute_capability
    return None


KNOWN_GPUS = [
    GPUInfo(name="A10", memory=24, compute_capability=(8, 6)),
    GPUInfo(name="A40", memory=48, compute_capability=(8, 6)),
    GPUInfo(name="A100", memory=40, compute_capability=(8, 0)),
    GPUInfo(name="A100", memory=80, compute_capability=(8, 0)),
    GPUInfo(name="A10G", memory=24, compute_capability=(8, 6)),
    GPUInfo(name="A4000", memory=16, compute_capability=(8, 6)),
    GPUInfo(name="A4500", memory=20, compute_capability=(8, 6)),
    GPUInfo(name="A5000", memory=24, compute_capability=(8, 6)),
    GPUInfo(name="A6000", memory=48, compute_capability=(8, 6)),
    GPUInfo(name="H100", memory=80, compute_capability=(9, 0)),
    GPUInfo(name="L4", memory=24, compute_capability=(8, 9)),
    GPUInfo(name="L40", memory=48, compute_capability=(8, 9)),
    GPUInfo(name="P100", memory=16, compute_capability=(6, 0)),
    GPUInfo(name="RTX3060", memory=8, compute_capability=(8, 6)),
    GPUInfo(name="RTX3060", memory=12, compute_capability=(8, 6)),
    GPUInfo(name="RTX3060Ti", memory=8, compute_capability=(8, 6)),
    GPUInfo(name="RTX3070Ti", memory=8, compute_capability=(8, 6)),
    GPUInfo(name="RTX3080", memory=10, compute_capability=(8, 6)),
    GPUInfo(name="RTX3080Ti", memory=12, compute_capability=(8, 6)),
    GPUInfo(name="RTX3090", memory=24, compute_capability=(8, 6)),
    GPUInfo(name="RTX4090", memory=24, compute_capability=(8, 9)),
    GPUInfo(name="RTX6000", memory=24, compute_capability=(7, 5)),
    GPUInfo(name="T4", memory=16, compute_capability=(7, 5)),
    GPUInfo(name="V100", memory=16, compute_capability=(7, 0)),
    GPUInfo(name="V100", memory=32, compute_capability=(7, 0)),
]



from gpuhunt.providers.vastai import kilo, normalize_gpu_memory


class TestGPU:
    def test_normalize_known(self):
        assert normalize_gpu_memory("A100", 78 * kilo) == 80

    def test_normalize_unknown(self):
        assert normalize_gpu_memory("X1000", 78 * kilo + 10) == 78



import datetime
import logging
import re
import time
from collections import defaultdict
from typing import Dict, List, Literal, Optional, TypedDict

import bs4
import jwt
import requests

from gpuhunt import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
API_URL = "api.ai.nebius.cloud"
COMPUTE_SERVICE_ID = "bfa2pas77ftg9h3f2djj"
GPU_NAME_PLATFORM = {
    "A100": "gpu-standard-v3",
    "H100 PCIe": "standard-v3-h100-pcie",
    "Hopper H100 SXM (Type A)": "gpu-h100",
    "Hopper H100 SXM (Type B)": "gpu-h100-b",
    "L4": "standard-v3-l4",
    "L40": "standard-v3-l40",
    None: "standard-v2",
}
GPU_PLATFORMS = {
    "gpu-standard-v3": [
        [28, 119.0, 1, "A100", 80.0],
        [56, 238.0, 2, "A100", 80.0],
        [112, 476.0, 4, "A100", 80.0],
        [224, 952.0, 8, "A100", 80.0],
    ],
    "gpu-h100": [
        [20, 160.0, 1, "H100", 80.0],
        [40, 320.0, 2, "H100", 80.0],
        [80, 640.0, 4, "H100", 80.0],
        [160, 1280.0, 8, "H100", 80.0],
    ],
    "gpu-h100-b": [
        [20, 160.0, 1, "H100", 80.0],
        [40, 320.0, 2, "H100", 80.0],
        [80, 640.0, 4, "H100", 80.0],
        [160, 1280.0, 8, "H100", 80.0],
    ],
    "standard-v3-h100-pcie": [
        [24, 96.0, 1, "H100", 80.0],
        [48, 192.0, 2, "H100", 80.0],
    ],
    "standard-v3-l4": [
        [4, 16.0, 1, "L4", 24.0],
        [8, 32.0, 1, "L4", 24.0],
        [12, 48.0, 1, "L4", 24.0],
        [16, 64.0, 1, "L4", 24.0],
        [24, 96.0, 1, "L4", 24.0],
        [24, 96.0, 2, "L4", 24.0],
        [48, 192.0, 2, "L4", 24.0],
    ],
    "standard-v3-l40": [
        [8, 32.0, 1, "L40", 48.0],
        [12, 48.0, 1, "L40", 48.0],
        [16, 64.0, 1, "L40", 48.0],
        [24, 96.0, 1, "L40", 48.0],
        [48, 192.0, 2, "L40", 48.0],
    ],
}
CPU_PLATFORMS = {
    "standard-v2": {
        "cpus": [
            2,
            4,
            6,
            8,
            10,
            12,
            14,
            16,
            20,
            24,
            28,
            32,
            36,
            40,
            44,
            48,
            52,
            56,
            60,
        ],
        "ratios": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
    }
}


class NebiusProvider(AbstractProvider):
    NAME = "nebius"

    def __init__(self, service_account: "ServiceAccount"):
        self.api_client = NebiusAPIClient(service_account)

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> List[RawCatalogItem]:
        zone = self.api_client.compute_zones_list()[0]["id"]
        skus = []
        page_token = None
        logger.info("Fetching SKUs")
        while True:
            page = self.api_client.billing_skus_list(
                filter=f'serviceId="{COMPUTE_SERVICE_ID}"', page_token=page_token
            )
            skus += page["skus"]
            page_token = page.get("nextPageToken")
            if page_token is None:
                break
        platform_resources = self.aggregate_skus(skus)
        offers = self.get_gpu_platforms(zone, platform_resources)
        offers += self.get_cpu_platforms(zone, platform_resources)
        return sorted(offers, key=lambda i: i.price)

    @staticmethod
    def get_gpu_platforms(
        zone: str, platform_resources: "PlatformResourcePrice"
    ) -> List[RawCatalogItem]:
        items = []
        for platform, presets in GPU_PLATFORMS.items():
            prices = platform_resources[platform]
            for cpu, memory, gpu_count, gpu_name, gpu_memory in presets:
                items.append(
                    RawCatalogItem(
                        instance_name=platform,
                        location=zone,
                        price=round(
                            cpu * prices["cpu"]
                            + memory * prices["ram"]
                            + gpu_count * prices["gpu"],
                            5,
                        ),
                        cpu=cpu,
                        memory=memory,
                        gpu_count=gpu_count,
                        gpu_name=gpu_name,
                        gpu_memory=gpu_memory,
                        spot=False,
                        disk_size=None,
                    )
                )
        return items

    @staticmethod
    def get_cpu_platforms(
        zone: str, platform_resources: "PlatformResourcePrice"
    ) -> List[RawCatalogItem]:
        items = []
        for platform, limits in CPU_PLATFORMS.items():
            prices = platform_resources[platform]
            for ratio in limits["ratios"]:
                for cpu in limits["cpus"]:
                    items.append(
                        RawCatalogItem(
                            instance_name=platform,
                            location=zone,
                            price=round(cpu * prices["cpu"] + cpu * ratio * prices["ram"], 5),
                            cpu=cpu,
                            memory=cpu * ratio,
                            gpu_count=0,
                            gpu_name=None,
                            gpu_memory=None,
                            spot=False,
                            disk_size=None,
                        )
                    )
        return items

    @staticmethod
    def parse_gpu_platforms(raw_html: str) -> Dict[str, List[List]]:
        """Parse GPU platforms from Nebius docs.

        Returns:
            Dict of platform name to a list of presets.
            Each preset contains: [cpu, memory, gpu_count, gpu_name, gpu_memory]
        """
        soup = bs4.BeautifulSoup(raw_html, "html.parser")
        configs = soup.find("h2", id="config").find_next_sibling("ul").find_all("li")
        platforms = {}
        for li in configs:
            platform = li.find("p").find("code").text
            gpu_name = re.search(r" ([A-Z]+\d+)[ \n]", li.find("p").text).group(1)
            items = []
            for tr in li.find("tbody").find_all("tr"):
                tds = tr.find_all("td")
                gpu_count = int(tds[0].text.strip(" *"))
                items.append(
                    [
                        int(tds[2].text),
                        float(tds[3].text),
                        gpu_count,
                        gpu_name,
                        int(tds[1].text) / gpu_count,
                    ]
                )
            platforms[platform] = items
        return platforms

    @staticmethod
    def parse_cpu_platforms(raw_html: str) -> Dict[str, Dict[str, List]]:
        """Parse CPU platforms from Nebius docs.

        Returns:
            Dict of platform name to Dict of resource name to a list of values.
        """
        soup = bs4.BeautifulSoup(raw_html, "html.parser")
        configs = (
            soup.find(
                "p",
                string=re.compile(
                    r"The computing resources may have the following configurations:"
                ),
            )
            .find_next_sibling("ul")
            .find_all("li")
        )
        platforms = {}
        for li in configs:
            platform = li.find("p").find("code").text
            tds = li.find("tbody").find("td", string="100%").find_next_siblings("td")
            platforms[platform] = {
                "cpus": [int(i) for i in tds[0].text.translate({"\n": "", " ": ""}).split(",")],
                "ratios": [
                    float(i) for i in tds[1].text.translate({"\n": "", " ": ""}).split(",")
                ],
            }
        return platforms

    def aggregate_skus(self, skus: List[dict]) -> "PlatformResourcePrice":
        vm_resources = {
            "GPU": "gpu",
            "RAM": "ram",
            "100% vCPU": "cpu",
        }
        vm_name_re = re.compile(
            r"((?:Intel|AMD) .+?)(?: with Nvidia (.+))?"
            rf"\. ({'|'.join(vm_resources)})(?: — (preemptible).*)?$"
        )
        platform_resources = defaultdict(dict)
        for sku in skus:
            if (r := vm_name_re.match(sku["name"])) is None:
                continue  # storage, images, snapshots, infiniband
            cpu_name, gpu_name, resource_name, spot = r.groups()
            if spot is not None:
                continue
            if gpu_name not in GPU_NAME_PLATFORM:
                logger.warning("Unknown GPU name: %s", gpu_name)
                continue
            platform_resources[GPU_NAME_PLATFORM[gpu_name]][
                vm_resources[resource_name]
            ] = self.get_sku_price(sku["pricingVersions"])

        return platform_resources

    @staticmethod
    def get_sku_price(pricing_versions: List[dict]) -> Optional[float]:
        now = datetime.datetime.now(datetime.timezone.utc)
        price = None
        for version in sorted(pricing_versions, key=lambda p: p["effectiveTime"]):
            # I guess it's the price for on-demand instances
            if version["type"] != "STREET_PRICE":
                continue
            if datetime.datetime.fromisoformat(version["effectiveTime"]) > now:
                break
            # I guess we should take the first pricing expression
            price = float(version["pricingExpressions"][0]["rates"][0]["unitPrice"])
        return price


class NebiusAPIClient:
    # reference: https://nebius.ai/docs/api-design-guide/
    def __init__(self, service_account: "ServiceAccount"):
        self._service_account = service_account
        self._s = requests.Session()
        self._expires_at = 0

    def get_token(self):
        now = int(time.time())
        if now + 60 < self._expires_at:
            return
        logger.debug("Refreshing IAM token")
        expires_at = now + 3600
        payload = {
            "aud": self.url("iam", "/tokens"),
            "iss": self._service_account["service_account_id"],
            "iat": now,
            "exp": expires_at,
        }
        jwt_token = jwt.encode(
            payload,
            self._service_account["private_key"],
            algorithm="PS256",
            headers={"kid": self._service_account["id"]},
        )

        resp = requests.post(payload["aud"], json={"jwt": jwt_token})
        resp.raise_for_status()
        iam_token = resp.json()["iamToken"]
        self._s.headers["Authorization"] = f"Bearer {iam_token}"
        self._expires_at = expires_at

    def billing_skus_list(
        self,
        filter: Optional[str] = None,
        page_size: Optional[int] = 1000,
        page_token: Optional[str] = None,
    ) -> "BillingSkusListResponse":
        logger.debug("Fetching SKUs")
        params = {
            "currency": "USD",
            "pageSize": page_size,
        }
        if filter is not None:
            params["filter"] = filter
        if page_token is not None:
            params["pageToken"] = page_token
        self.get_token()
        resp = self._s.get(self.url("billing", "/skus"), params=params)
        resp.raise_for_status()
        return resp.json()

    def compute_zones_list(self) -> List[dict]:
        logger.debug("Fetching compute zones")
        self.get_token()
        resp = self._s.get(self.url("compute", "/zones"))
        resp.raise_for_status()
        return resp.json()["zones"]

    def url(self, service: str, path: str, version="v1") -> str:
        return f"https://{service}.{API_URL.rstrip('/')}/{service}/{version}/{path.lstrip('/')}"


class ServiceAccount(TypedDict):
    id: str
    service_account_id: str
    created_at: str
    key_algorithm: str
    public_key: str
    private_key: str


class BillingSkusListResponse(TypedDict):
    skus: List[dict]
    nextPageToken: Optional[str]


PlatformResourcePrice = Dict[str, Dict[Literal["cpu", "ram", "gpu"], float]]
